#include "OpenCLManager.h"
#include "OpenCLErrorHandler.h"
#include <stdio.h>
#include <stdlib.h>   //  für  malloc, free

cl_int loadAndCompileProgram(const char* fileName, 
                             cl_context context,
                             cl_program *program ) {
	cl_int err;
	cl_int res = CL_TRUE;
	FILE* f = fopen(fileName, "rb");
	if (f == NULL) {
		return CL_FALSE;
	}
    fseek(f,0,SEEK_END);
    int fSize = ftell(f);
    fseek(f,0,SEEK_SET);
    {
        cl_char codeString[fSize + 1];
        fread(codeString, 1, fSize, f);
        codeString[fSize] = 0; //null terminate it!
        fclose(f);
        printf("%s\n\n", codeString);
        cl_char* refString = &codeString[0];
        *program = clCreateProgramWithSource
	            (context, 1, (const char**) &refString, NULL, &err);
	    printf("result of createProgram: %d\n", err);
	}
	return err == CL_SUCCESS;
}

int initializeOpenCL (gpuModel *gpu) {
	
	cl_context_properties properties[3];
    cl_uint nroOfPlatforms;
    cl_platform_id platformId;
    cl_device_id deviceId;
    cl_uint nroOfDevices;
    cl_int err;
    
	if (clGetPlatformIDs(1, &platformId, &nroOfPlatforms) != CL_SUCCESS) {
		printf("Unable to get platform id\n");
		return 1;
	}
	
	if (clGetDeviceIDs(platformId, CL_DEVICE_TYPE_GPU,
	                   1, &deviceId, &nroOfDevices) != CL_SUCCESS) {
		printf("Unable to get device id\n");
		return 1;
	}
	
	properties[0] = CL_CONTEXT_PLATFORM;
	properties[1] = (cl_context_properties) platformId;
	properties[2] = 0;
	
	gpu -> context = clCreateContext(properties, 1,
	                          &deviceId, NULL, NULL, &err);
    //  clCreateCommandQueue  ist in den Headerfiles als deprecated
    //  markiert, wird für OpenCL 1.2 aber gebraucht, da das neuere
    //  clCreateCommandQueueWithProperties  nicht zur Verfügung
    //  steht.!
	gpu -> commandQueue = clCreateCommandQueue(gpu -> context, deviceId, 0, &err);
	
    loadAndCompileProgram("res/kernel.cl", gpu -> context , &gpu -> program );
    
    err = clBuildProgram(gpu -> program, 0, NULL, NULL, NULL, NULL);
    if (err != CL_SUCCESS) {
		printf("error building program:  ");
		checkError(err);
		// Abholen der Fehlermeldung: Das ist durchaus tricky
		if (err != CL_SUCCESS) {
			char* errorMessage;
			int errFlag;
			size_t msgSize;
			errFlag = clGetProgramBuildInfo(gpu -> program,
			                                deviceId,
			                                CL_PROGRAM_BUILD_LOG,
			                                0,
			                                NULL,
			                                &msgSize);
			char msg [msgSize + 10];
			
			errFlag = clGetProgramBuildInfo(gpu -> program,
			                                deviceId,
			                                CL_PROGRAM_BUILD_LOG,
			                                msgSize + 10,
			                                &msg,
			                                &msgSize);
			printf("%s\n\n", msg);                          
		}
		return 1;
	}
	
	gpu -> kernel = clCreateKernel(gpu -> program, "fraktal", &err);
	checkError(err);
	
	return 0;
}

void createImage(gpuModel *gpu) {
    cl_image_format format;
    cl_image_desc imageDescriptor;
    cl_int err;
    int size = gpu -> imageSize * gpu -> imageSize * 4;
    gpu -> imageBuffer = malloc (size * sizeof (unsigned char));
	gpu -> output = NULL;
	format.image_channel_order = CL_RGBA;
	format.image_channel_data_type = CL_UNSIGNED_INT8;
	imageDescriptor.image_type = CL_MEM_OBJECT_IMAGE2D;
	imageDescriptor.image_width =  gpu -> imageSize;
	imageDescriptor.image_height = gpu -> imageSize;
	imageDescriptor.image_depth = 0;  // not used for 2d image
	imageDescriptor.image_array_size = 0;
	imageDescriptor.image_row_pitch = 0;
	imageDescriptor.image_slice_pitch = 0;
	imageDescriptor.num_mip_levels = 0;
	imageDescriptor.num_samples = 0;
	imageDescriptor.buffer = NULL; 
	gpu -> output = clCreateImage(gpu -> context,
	                              CL_MEM_WRITE_ONLY,
	                              &format,
	                              &imageDescriptor,
	                              NULL,
	                              &err);
	checkError(err);

}



void performComputation(applicationModel *model) {
	windowElements *gui = &model -> gui;
	gpuModel *gpu = &model -> gpu;
	
    cl_int err;
	size_t origin [] = {0, 0, 0};
	size_t region [] = {gpu -> imageSize, gpu -> imageSize, 1};
	double centerPt [] = { 0.0, 0.0 };
	int w = gpu -> imageSize,
	    h = gpu -> imageSize;
	double alpha = gtk_adjustment_get_value(gui -> adjustmentAlpha),
	       max = gtk_adjustment_get_value(gui -> adjustmentZoom)/10.0;
	double min = -max;
	double scale = (max - min)/gpu->imageSize;
	size_t global [] = { gpu -> imageSize, gpu -> imageSize };
	size_t local [] = { 16, 16 };  // das ist bereits die MAX_WORK_GROUP_SZIE
/*
	err = clEnqueueWriteImage(gpu -> commandQueue, (gpu -> output),
	                          CL_TRUE,
	                          origin,
	                          region,
	                          0, // imageDescriptor.image_width*sizeof(int),
	                          0, // imageDescriptor.image_width*imageDescriptor.image_height*sizeof(int),
	                          (*gpu).imageBuffer, 0, NULL, NULL);
	checkErrorMsg(err, "clEnqueueWriteImage");  */
	
	err = clSetKernelArg(gpu -> kernel, 0, sizeof(cl_mem), &(gpu -> output));
	checkErrorMsg(err, "SetKernelArg 0");
	err = clSetKernelArg(gpu -> kernel, 1, sizeof(int), &w);
	checkErrorMsg(err, "SetKernelArg 1");
	err = clSetKernelArg(gpu -> kernel, 2, sizeof(int), &h);
	checkErrorMsg(err, "SetKernelArg 2");
	err = clSetKernelArg(gpu -> kernel, 3, 2*sizeof(double), &centerPt);
	checkErrorMsg(err, "SetKernelArg 3");
	err = clSetKernelArg(gpu -> kernel, 4, sizeof(double), &scale);
	checkErrorMsg(err, "SetKernelArg 4");
	err = clSetKernelArg(gpu -> kernel, 5, sizeof(double), &alpha);
	checkErrorMsg(err, "SetKernelArg 5");
	
	clEnqueueNDRangeKernel(gpu -> commandQueue,
	                       gpu -> kernel,
	                       2,           // dimension
	                       NULL,        // global work offset {0, 0}
	                       global,      // global work size
	                       local,       // local work size (set automatically)
	                       0,           // num events in wait list
	                       NULL,        // event wait list
	                       NULL);       // event
	checkErrorMsg(err, "clEnqueueNDRangeKernel");
	clFinish(gpu -> commandQueue);
	
	err = clEnqueueReadImage(gpu -> commandQueue, (gpu -> output),
	                          CL_TRUE,
	                          origin,
	                          region,
	                          0, // imageDescriptor.image_width*sizeof(int),
	                          0, // imageDescriptor.image_width*imageDescriptor.image_height*sizeof(int),
	                          (*gpu).imageBuffer, 0, NULL, NULL);
	checkErrorMsg(err, "clEnqueueReadImage");
}

void freeImage(gpuModel *gpu) {
    free(gpu -> imageBuffer);
    gpu -> imageBuffer = (char) 0;
    clReleaseMemObject(gpu -> output);
    gpu -> output = NULL;
}

void freeResources (gpuModel *gpu) {
	printf("freeResources 0\n"); 
	if (gpu -> output != NULL) {
        clReleaseMemObject(gpu -> output);
        gpu -> output = NULL;
	}
	if (gpu -> imageBuffer != (char) 0) {
		free(gpu -> imageBuffer);
        gpu -> imageBuffer = (char) 0;
    }
    clReleaseProgram(gpu -> program);
	printf("freeResources 1\n");
    clReleaseKernel(gpu -> kernel);
	printf("freeResources 2\n");
    clReleaseCommandQueue(gpu -> commandQueue);
	printf("freeResources 3\n");
    clReleaseContext(gpu -> context);
	printf("freeResources 4\n");
}
